/** \file CahnHilliardNavierStokes_.h */

#ifndef CAHN_HILLIARD_NAVIER_STOKES_H
#define CAHN_HILLIARD_NAVIER_STOKES_H

// coupling structures
#include "CouplingIterationInterface.h"
#include "CouplingTimeInterface.h"
#include "CouplingProblemStat.h"

#include "CahnHilliard_.h"

// structures for local refinement
#include "Refinement.h"
#include "MeshFunction_Level.h"

#include "POperators.h"

using namespace AMDiS;








class Y : public AbstractFunction<double, WorldVector<double> >
{
public:
  Y()
    : AbstractFunction<double, WorldVector<double> >(1)  {};
  double operator()(const WorldVector<double>& x) const  {
    return x[1];
  }
};







/**
  * \ingroup Problem 
  *
  * \brief
  */
template<typename CH_Type = CahnHilliard_,
	 typename NS_Type = NavierStokes_TaylorHood>
class CahnHilliardNavierStokes_ : public CouplingIterationInterface,
                                 public CouplingTimeInterface,
                                 public CouplingProblemStat
{
public:

  CahnHilliardNavierStokes_(std::string name_, CH_Type *chProb_, NS_Type *nsProb_)
  : CouplingProblemStat(name_),
    chProb(chProb_),
    nsProb(nsProb_),
    refFunction(NULL),
    refinement(NULL),
    sigma(0.0),
    surfaceTension(0.0)
  {
    dow = Global::getGeo(WORLD);

    Parameters::get(name + "->sigma", sigma);
    surfaceTension = sigma/chProb->getEpsilon() * 3.0/2.0/sqrt(2.0);
    

  }


  ~CahnHilliardNavierStokes_() {
    if (refFunction != NULL)
      delete refFunction;
    if (refinement != NULL)
      delete refinement;
  }


  /**
   * Add the problems to the iterationInterface, timeInterface and couplingProblemStat.
   * As a consequence all problem can be initialized one after another and in the
   * adaption loop they are solved in rotation.
   * At the end the problems are filled with operators and coupling operators as well as
   * boundary conditions are added.
   *
   * In the adaption loop the problems are solved the same order as they are added to the
   * iterationInterface in this method. This order can be changed manually in the oneIteration
   * method.
   **/
  void initialize(AdaptInfo *adaptInfo)
  {
    for (size_t i = 0; i < chProb->getNumProblems(); i++)
      addProblem(chProb->getProblem(i));
    for (size_t i = 0; i < nsProb->getNumProblems(); i++)
      addProblem(nsProb->getProblem(i));
    
    addIterationInterface(chProb);
    addIterationInterface(nsProb);
   // setSolveProblem("ch",false);
 

    addTimeInterface(chProb);
    addTimeInterface(nsProb);

    CouplingProblemStat::initialize(INIT_ALL);
    dim = getMesh()->getDim();
    fillCouplingOperators();
    // fillOperators and fillBoundaryConditions for chProb and nsProb
    nsProb->initTimeInterface();
    chProb->initTimeInterface();
    
    y = new DOFVector<double>(chProb->getFeSpace(0), "volume");
    yc1 = new DOFVector<double>(chProb->getFeSpace(0), "volume");
    yc2 = new DOFVector<double>(chProb->getFeSpace(0), "volume");
    output = fopen( "output_data", "w" );
    fprintf(output, "time u_b center_old\n");
    fclose(output);
    u_b=0;
//     fillCouplingBoundaryConditions();
  }

  
  Flag oneIteration(AdaptInfo *adaptInfo, Flag toDo)
  { FUNCNAME("CouplingIterationInterface::oneIteration()");
    Flag flag = CouplingIterationInterface::oneIteration(adaptInfo, toDo);
    
    //chProb->solveInitialProblem(adaptInfo);
    
    return flag;
  };
  

  /**
   * In this method the operators responsible for the coupling of all problems are defined
   * and added to the problem. The access to the problems is through the baseBroblem->getProblem()
   * method. All baseProblem provide FiniteElemSpaces and solution vectors.
   **/
  void fillCouplingOperators()
  { FUNCNAME("CahnHilliardNavierStokes::fillCouplingOperators()");
    MSG("CahnHilliardNavierStokes::fillCouplingOperators()");

    for (size_t i = 0; i < dow; i++) {
      // < nu * d_i(c) , theta >
      Operator *opNuGradC = new Operator(nsProb->getFeSpace(i), chProb->getFeSpace(0));
      opNuGradC->addTerm(new VecAndPartialDerivative_ZOT(chProb->getSolution()->getDOFVector(0), chProb->getSolution()->getDOFVector(1), i));
      nsProb->getProblem(0)->addVectorOperator(opNuGradC, i, &surfaceTension, &surfaceTension);
    }

    // < v * grad(c) , theta >
    Operator *opVGradC = new Operator(chProb->getFeSpace(0), chProb->getFeSpace(1));
    opVGradC->addTerm(new WorldVector_FOT(nsProb->getVelocity(), 1.0), GRD_PHI);
    chProb->getProblem()->addMatrixOperator(opVGradC, 1, 1,  chProb->getTau());
   /**/ 
  }
  

  /**
   * Solves the initial problems.Before this is done the mesh is refined by the Refinement-class
   * This provides local refinement functions for phaseField-refinement or signed-distance-refinement.
   * See MeshFunction_Level.h for some details.
   **/ 
  void solveInitialProblem(AdaptInfo *adaptInfo)
  {
    // meshFunction for klocal refinement around the interface of the phasefield-DOFVector
    refFunction = new PhaseFieldRefinement(chProb->getMesh());

    if (chProb->getDoubleWellType() == 0) {
      refinement = new RefinementLevelDOF(
	chProb->getFeSpace(),
	refFunction,
	chProb->getSolution()->getDOFVector(1)); // phaseField-DOFVector in [0,1]
    } else {
      refinement = new RefinementLevelDOF(
	chProb->getFeSpace(),
	refFunction,
	new PhaseDOFView<double>(chProb->getSolution()->getDOFVector(1))); // phaseField-DOFVector in [-1,1]
    }

    // initial refinement
    refinement->refine(0);

    for (int i = 0; i < 3; i++) {
      chProb->solveInitialProblem(adaptInfo); 	// update phaseField-DOFVector
      refinement->refine((i < 4 ? 4 : 10));	// do k steps of local refinement
    }

    // solve all initial problems of the problems added to the CouplingTimeInterface
    CouplingTimeInterface::solveInitialProblem(adaptInfo);
  }


  /**
   * Called at the end of each timestep.
   * If the phase-field has changed the mesh is updated by the refinement-method
   **/
  void closeTimestep(AdaptInfo *adaptInfo)
  {
    CouplingTimeInterface::closeTimestep(adaptInfo);

    
    
    Y yFun;
    y->interpol(&yFun);

    DOFVector<double>::Iterator phiIt(dynamic_cast<DOFIndexed<double>*>(const_cast<DOFVector<double>*>(chProb->getSolution()->getDOFVector(1))), USED_DOFS);
    DOFVector<double>::Iterator yIt(dynamic_cast<DOFIndexed<double>*>(y), USED_DOFS);
    DOFVector<double>::Iterator yc1It(dynamic_cast<DOFIndexed<double>*>(yc1), USED_DOFS);
    DOFVector<double>::Iterator yc2It(dynamic_cast<DOFIndexed<double>*>(yc2), USED_DOFS);

    for (phiIt.reset(), yIt.reset(),yc1It.reset(),yc2It.reset();
         !phiIt.end();
         ++phiIt, ++yIt,++yc1It,++yc2It){
                *yc2It = ((1.0)+*phiIt)/2.0;
                *yc1It = *yc2It * *yIt;
	}

        if (adaptInfo->getTimestepNumber() <= 1) center_old = (yc1->Int())/(yc2->Int());
        u_b = - (center_old - (yc1->Int())/(yc2->Int()) ) / adaptInfo->getTimestep();
	center_old = (yc1->Int())/(yc2->Int());
    
    output = fopen( "output_data", "a" );
    fprintf(output, "%f, %f, %f\n",
                adaptInfo->getTime(),
                u_b,
                center_old);
    fclose(output);

//chProb->solveInitialProblem(adaptInfo);
    
    refinement->refine(2);
  }
  
  
  
  

protected:

  CH_Type *chProb;	// CahnHilliard baseProblem
  NS_Type *nsProb;	// Navier-Stokes baseProblem

  PhaseFieldRefinement* refFunction;
  RefinementLevelDOF *refinement;

  unsigned dim;		// dimension of the mesh
  unsigned dow;		// dimension of the world

  double sigma;		// coupling parameter to calculate the surface tension
  double surfaceTension;// := sigma/epsilon
  double u_b, center_old;
  
  FILE *output;
  
  DOFVector<double> *y, *yc1, *yc2;
};

#endif // CAHN_HILLIARD_NAVIER_STOKES_H
